/************************************************************************************/
/* Jakub Skowski H5ISI boost::operators											*/
/* Biblioteka pomaga programicie w implementacji przecionych operatorw klas		*/
/* przez implementacj szeregu klas bazowych specjalizujcych sie w wykonaniu		*/
/* charakterystycznych operacj. Wymaga ona jednak od uytkownika przecienia		*/ 
/* podstawowego dla danej klasy bazowej operatora.									*/
/* np. Za pomoc operatora "<" mona zapisa operatory: "<=", ">=" oraz ">".		*/
/* Aby znacznie nie powiksza zbioru operatorw przecionych dla klasy mamy		*/
/* moliwo wyprowadzenia naszej klasy z szablonowej klasy bazowej					*/
/* "less_than_comparable". Teraz musimy juz zaimplementowa tylko operator< - reszt*/
/* implementuje nasza klasa bazowa													*/
/************************************************************************************/

/*    Przykad    */
#include <iostream>
#include <boost/operators.hpp>

class some_class: boost::less_than_comparable<some_class>{

	int data;

public:
	some_class(int num):data(num){
	}
	bool less_than(const some_class& lhs)const{
		return data<lhs.data;
	}
//	friend bool operator<(const some_class& a, const some_class& b){
//		return a.data<b.data;
//	}
};

bool operator<(const some_class& lhs, const some_class& rhs){
	return lhs.less_than(rhs);
}
int main(){

	some_class ob1(3);
	some_class ob2(5);

	std::cout<<(ob1<ob2);
	std::cout<<(ob1>ob2);
	std::cout<<(ob1<=ob2);
	std::cout<<(ob1>=ob2);
	return 0;
}
/********************************************************************************************************/
/*		Wynikiem dziaania programu jest oczywicie: 1010 												*/
/* Zamiast implementacji funkcji less_than moemy doda do metod klasy zaprzyjanion funkcj operator<	*/
/*  (zakomentowane), co jest jak najbardziej poprawn implementacj.									*/
/********************************************************************************************************/
/* Poza less_than_comparable do dyspozycji mamy:														*/
/* klasa bazowa				opearacje zaimplementowane:				 wymagane od klasy pochodnej:		*/
/* equality_comparable		bool operator==(const T&, const T&)		bool operator==(const T&, const T&)	*/
/* 							bool operator!=(const T&, const T&)											*/
/* addable					T	 operator+ (const T&, const T&)		T operator+=(const T&)				*/
/* 							T	 operator+=(const T&)													*/
/* subtractable				T	 operator- (const T&, const T&)		T operator-=(const T&)				*/
/* 							T	 operator-=(const T&)													*/
/* orable					T	 operator| (const T&, const T&)		T operator|=(const T&, const T&)	*/
/* 							T	 operator|=(const T&, const T&)											*/
/* andable					T	 operator& (const T&, const T&)		T operator&=(const T&, const T&)	*/
/* 							T	 operator&=(const T&, const T&)											*/
/* incrementable			T&	 operator++(T&)						T& operator++(T&)					*/
/* 							T	 operator++(T&, int)													*/
/* decrementable			T&	 operator--(T&)						T& operator--(T&)					*/
/* 							T	 operator--(T&, int)													*/
/* equivalent				T&	 operator< (T&)						T& operator<(const T&, const &T)	*/
/* 							T	 operator++(T&, int)													*/
/* dereferenceable			R	 operator->() const					R* operator*() const				*/
/* 							R	 operator* () const														*/
/* 							(T - typ operandu, R - typ referencji T, P -typ wskanika T)				*/
/* indexable				R	 operator[](D) const				R operator+	(const T&, D)			*/
/* 							R	 operator+ (const T&, D)												*/
/* 							(D - typ wartoci indeksujcych)											*/
/********************************************************************************************************/
/* To s tylko podstawowe klasy, istnieje duo wicej gromadzcych w sobie takie operacje jak xor lub   */ 
/* mod. Istniej take klasy grupujce w/w klasy.														*/
/* np:																									*/
/* totaly_ordered zawiera: less_than_comparable i equality_comparable									*/
/* additive:				addable i subtractable														*/
/* itp, itd.....  wicej szczegw: http://www.boost.org/												*/
/*																										*/
/* Moemy dizedziczy wielobazowo:																		*/
/* class some_class: boost::less_than_comparable<some_class>,											*/
/* 			      boost::equivalent<some_class>{														*/
/* }																									*/
/*																										*/
/* Naley zwrci uwag na rozrastanie si klas. Niektre kompilatory nieradz sobie z optymalnym		*/
/* dziedziczeniem wielobazowym, nadmiernie zwikszajc rozmiar obiektw klas pochodnych kopiujc		*/
/* wielokrotnie pocztkowe klasy bazowe. Problem ten rozwizany jest za pomoc acucha dziedziczenia	*/ 
/* jednobazowego. Kady szablon funkcji bazowej w boost::operators moe przyjmowa opcjonalnie poza		*/
/* klas pochodn, klas bazow po ktrej ma dziedziczy. W ten sposb z klas bazowych moemy utwoy	*/
/* acuch w ktrym kada klasa bdzie wskazywaa poprzednika w acuchu dziedziczenia.					*/
/* Dobrym przykladem s typy atytmetyczne, od ktrych mona wymaga implementacji wielu operatorw.		*/
/* ******************************************************************************************************/

/*		Przykad		*/

#include<iostream>
#include<boost/operators.hpp>

template<class T>
class some_class2:
	boost::addable<some_class2<T>, 
		boost::subtractable<some_class2<T>,
			boost::less_than_comparable<some_class2<T>,
				boost::equality_comparable<some_class2<T>>
	>>>
	{
	public:
		some_class2():t(0){}
		some_class2(T t_):t(t_){}
		T get(){
			return t;
		}

		// dla less_than_comparable
		friend bool operator<(const some_class2<T>& lhs, const some_class2<T>& rhs){
			return lhs.t<rhs.t;
		}
		
		//dla equality_comparable
		friend bool operator==(const some_class2<T>& lhs, const some_class2<T>& rhs){
			return lhs.t==rhs.t;
		}

		//dla addable
		some_class2<T>& operator+=(const some_class2<T>& other){
			t+=other.t;
			return *this;
		}

		//dla subtractable
		some_class2<T>& operator-=(const some_class2<T>& other){
			t-=other.t;
			return *this;
		}

	private:
		T t;
	};

/*********************************************************************************************************/
/* Taki zapis deklaracji klasy, gdzie jako 2 argument szablonu klasy podajemy klas bazow zapewnia		 */
/* odpowiedni acuch dziedziczenia oraz zawarcie w nim wszystkich niezbednych klas. Zauwamy take, e  */
/* zapisujc implementacj przecienia 4 operatorw, mamy dostp do 10, modyfikujc tylko deklaracj    */
/* klasy. Zmniejsza to moliwo popenienia bdu przez programist. Pamitajmy, i mamy dostp do typw*/ 
/* zoonych, co sprawia, e nasz deklaracj klasy mona sprowadzi do postaci:						 */
/*********************************************************************************************************/

template<class T>
class some_class2:
	boost::additive<some_class2<T>, 
		boost::totally_ordered<some_class2<T>>>

/*********************************************************************************************************/
/* Jak wida jest to rozwizanie duo krtsze co sprawia, e trdniej jest si pomyli.					 */
/* Operators posiada take implementacj klas zajmujcych si operatorami indeksowania i dereferencji	 */
/* przydatne gwnie w klasach zachowujcych si podobnie do iteratorw. Zasada dziaania analogiczna    */
/* do poprzednich przykadw.																			 */
/*********************************************************************************************************/